//=============================================================================================================================
//
// Copyright (c) 2015-2022 VisionStar Information Technology (Shanghai) Co., Ltd. All Rights Reserved.
// EasyAR is the registered trademark or trademark of VisionStar Information Technology (Shanghai) Co., Ltd in China
// and other countries for the augmented reality technology developed by VisionStar Information Technology (Shanghai) Co., Ltd.
//
//=============================================================================================================================

#import "helloar.h"
#import "MyCamera.h"
#import "BoxRenderer.h"
#import "BGRenderer.h"
#import <easyar/vector.oc.h>
#import <easyar/types.oc.h>
#import <easyar/buffer.oc.h>
#import <easyar/camera.oc.h>
#import <easyar/frame.oc.h>
#import <easyar/imagetracker.oc.h>
#import <easyar/imagetarget.oc.h>
#import <easyar/callbackscheduler.oc.h>
#include <OpenGLES/ES3/gl.h>
#import <GLKit/GLKView.h>
#import <easyar/dataflow.oc.h>
#import <easyar/cameraparameters.oc.h>
#import <easyar/image.oc.h>

BGRenderer * bgRenderer;
MyCamera * camera;
NSMutableArray<easyar_ImageTracker *> * trackers;
easyar_DelayedCallbackScheduler * scheduler;
BoxRenderer * boxRenderer;
easyar_InputFrameThrottler * throttler;
easyar_InputFrameToFeedbackFrameAdapter * i2FAdapter;
easyar_OutputFrameBuffer * outputFrameBuffer;
easyar_OutputFrameFork * outputFrameFork;
int previousInputFrameIndex = -1;

void loadFromImage(easyar_ImageTracker * tracker, NSString * path, NSString * name)
{
    easyar_ImageTarget * target = [easyar_ImageTarget createFromImageFile:path storageType:easyar_StorageType_Assets name:name uid:@"" meta:@"" scale:1];
    [tracker loadTarget:target callbackScheduler:scheduler callback:^(easyar_Target * target, bool status) {
        NSLog(@"load target (%d): %@ (%d)", status, [target name], [target runtimeID]);
    }];
}

void createScheduler()
{
    scheduler = [easyar_DelayedCallbackScheduler create];
}
easyar_DelayedCallbackScheduler * getScheduler()
{
    return scheduler;
}

void recreate_context()
{
    bgRenderer = nil;
    boxRenderer = nil;
    previousInputFrameIndex = -1;
    bgRenderer = [BGRenderer create];
    boxRenderer = [BoxRenderer create];
}

void initialize()
{
    recreate_context();

    throttler = [easyar_InputFrameThrottler create];
    i2FAdapter = [easyar_InputFrameToFeedbackFrameAdapter create];
    outputFrameBuffer = [easyar_OutputFrameBuffer create];
    outputFrameFork = [easyar_OutputFrameFork create:2];

    camera = [[MyCamera alloc] init];
    [camera open];
    
    easyar_ImageTracker * tracker = [easyar_ImageTracker createWithMode:easyar_ImageTrackerMode_PreferQuality];
    [tracker setSimultaneousNum:1];
    loadFromImage(tracker, @"namecard.jpg", @"namecard");
    trackers = [[NSMutableArray<easyar_ImageTracker *> alloc] init];
    [trackers addObject:tracker];
    
    [camera connect:[throttler input]];
    [[throttler output] connect:[i2FAdapter input]];
    [[i2FAdapter output] connect:[tracker feedbackFrameSink]];
    [[tracker outputFrameSource] connect:[outputFrameFork input]];
    [[outputFrameFork output:0] connect:[outputFrameBuffer input]];
    [[outputFrameFork output:1] connect:[i2FAdapter sideInput]];
    [[outputFrameBuffer signalOutput] connect:[throttler signalInput]];
}

void finalize()
{
    [trackers removeAllObjects];
    trackers = nil;
    boxRenderer = nil;
    bgRenderer = nil;
    camera = nil;
    throttler = nil;
    i2FAdapter = nil;
    outputFrameBuffer = nil;
    outputFrameFork = nil;
    scheduler = nil;
}

BOOL start()
{
    BOOL status = YES;
    if (camera != nil) {
        status &= [camera start];
    } else {
        status = NO;
    }
    for (easyar_ImageTracker * tracker in trackers) {
        status &= [tracker start];
    }
    return status;
}

void stop()
{
    for (easyar_ImageTracker * tracker in trackers) {
        [tracker stop];
    }
    if (camera != nil) {
        [camera stop];
    }
}
void render(int width, int height, int screenRotation)
{
    while ([scheduler runOne])
    {
    }

    glViewport(0, 0, width, height);
    glClearColor(0.f, 0.f, 0.f, 1.f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    easyar_OutputFrame * oFrame = [outputFrameBuffer peek];
    if (oFrame == nil){ return; }
    easyar_InputFrame * iFrame = [oFrame inputFrame];
    if (iFrame == nil) { return; }
    if (![iFrame hasCameraParameters]) { return; }
    easyar_CameraParameters * cameraParameters = [iFrame cameraParameters];
    float viewport_aspect_ratio = (float)width / (float)height;
    easyar_Matrix44F * projection = [cameraParameters projection:0.01f farPlane:1000.f viewportAspectRatio:viewport_aspect_ratio screenRotation:screenRotation combiningFlip:true manualHorizontalFlip:false];
    easyar_Matrix44F * imageProjection = [cameraParameters imageProjection:viewport_aspect_ratio screenRotation:screenRotation combiningFlip:true manualHorizontalFlip:false];
    easyar_Image * image = [iFrame image];

    if ([iFrame index] != previousInputFrameIndex) {
        [bgRenderer upload:[image format] width:[image width] height:[image height] pixelWidth:[image pixelWidth] pixelHeight:[image pixelHeight] bufferData:[[image buffer] data]];
        previousInputFrameIndex = [iFrame index];
    }
    [bgRenderer render:imageProjection];

    NSArray<easyar_FrameFilterResult *> * results = [oFrame results];
    
    for (int i = 0; i < [results count]; i++) {
        easyar_ImageTrackerResult * result = nil;
        easyar_FrameFilterResult * _result = results[i];
        result = ([_result isEqual:[NSNull null]])? nil: (easyar_ImageTrackerResult * )_result;
        if (result != nil) {
            for (easyar_TargetInstance * targetInstance in [result targetInstances]) {
                easyar_Target * target = [targetInstance target];
                easyar_ImageTarget * imagetarget = [target isKindOfClass:[easyar_ImageTarget class]] ? (easyar_ImageTarget *)target : nil;
                if (imagetarget == nil) {
                    continue;
                }
                easyar_Vec2F * scale = [easyar_Vec2F create:@[@([imagetarget scale]), @([imagetarget scale] / [imagetarget aspectRatio])]];
                [boxRenderer render:projection cameraview:[targetInstance pose] size:scale];
            }
        }
    }
}
